import java.util.*;

import finch.Finch;

public class SensorDriver {
	/** Total number of samples to take */
	public static final int numSamples = 10; 

	/**
	 * Interval between samples (note that there is also other overhead that
	 * extends this delay)
	 */
	public static final int sampleInterval = 1; // In ms

	
	/**
	 * Take a specified number of FinchSensor samples and place them into a
	 * map.
	 * 
	 * @param myFinch	A reference to a valid Finch object or null
	 * @param num		Number of samples to take
	 * @param interval	Interval in ms between samples (note that this 
	 * 						does not take into account the time required
	 * 						to take a sample.
	 * @param map		The map in which to insert the samples.  This map
	 * 						is keyed by the time of the sample
	 * 
	 * <dt><b>Conditions:</b>
     * <dd>PRE  -         The map already exists
     * <dd>POST -         A set of FinchSensor objects are inserted (as values) into the 
     *                      existing map.  FinchSensor is keyed with the sample time.
	 * </dt>
	 * 
	 */
	
	public static void getSamples(Finch myFinch, int num, int interval,
			Map<Long, FinchSensor> map){

		// Sensor values to store
		int light[];
		double acceleration[];
		double temperature;
		boolean obstacle[];
		
		if(myFinch != null){
			// Turn the nose red
			myFinch.setLED(255, 0, 0);
		}
		
		// Loop over the set of samples
		for(int i = 0; i < num; ++i){
			// Grab the sensor values
			if(myFinch != null){
				// Poll Finch for the data
				light = myFinch.getLightSensors();
				acceleration = myFinch.getAccelerations();
				temperature = myFinch.getTemperature();
				obstacle = myFinch.getObstacleSensors();
			}else{
				// Testing purposes only: fill in constant values
				light = new int[2];
				acceleration = new double[3];
				temperature = 100.0;
				obstacle = new boolean[2];
			}
			
			// Create a new FinchSensor and place it in the map
			Long sample_time = System.currentTimeMillis();
			map.put(sample_time,
					new FinchSensor(light, acceleration, obstacle, 
							temperature, sample_time));
			
			// Delay before the next sample
			if(myFinch != null){
				// Use the Finch-supplied sleep
				myFinch.sleep(sampleInterval);
			}else{
				// Otherwise, use a thread sleep (for testing purposes only)
				try{
					Thread.currentThread().sleep(interval);
				}catch(InterruptedException ie){
						//If this thread was interrupted by another thread 
				}
			}	
		}
		
		// Done sampling
		if(myFinch != null){
			// Turn the nose to black
			myFinch.setLED(0, 0, 0);
		}
	}

	/**
	 * Report a set of samples that have been stored in the map.  The report is
	 *  generated in key order
	 *  
     *              
	 * @param map 	Some map
	 * 
	 * <dt><b>Conditions:</b>
     * <dd>PRE  -         The map already exists
     * <dd>POST -         none
     * </dt>       
     * 
	 */
	
	public static <T, E> void reportSamples(Map<T, E> map) {
		// Pull out the keys that exist in the map
		Set<T> set = map.keySet();
		
		// Iterate over each key and display both the key and the value
		for (T key : set) {
			System.out.println("Key: " + key + "; Sample: " + map.get(key));
		}

	}

	/**
	 * Given a map and an array of keys, measure the amount of time (in ms)
	 * that is required to query all of the keys.
	 * 
	 * 
	 * @param map	A map from sample time (Long) to FinchSensor
	 * @param keys	An array of keys to access during the test (sample times)
	 * 
	 * <dt><b>Conditions:</b>
     * <dd>PRE  -         The map and key array already exist
     * <dd>POST -         none
	 * </dt>
	 * <P>
	 * 
	 * <DT><B>Algorithm: </B>
	 * <DD> Start a timer
	 * <DD> Loop over the entire set of keys to query and ask the map to retrieve
	 * the corresponding value
	 * <DD> Stop the timer.
	 * </DT>
	 * <P>
	 */
	public static void queryTest(Map<Long, FinchSensor> map, Long[] keys) {
		
		// Current time is the start time
		Long start_time = System.currentTimeMillis();
		
		// Loop over each key in the array and extract the associated value
		for (int i = 0; i < keys.length; ++i) {
			FinchSensor sensor = map.get(keys[i]);
		}
		
		// Current time is the end time
		Long end_time = System.currentTimeMillis();
		
		// Report the amount of time required for the entire set of queries
		System.out.println("Access Time: " + new Long(end_time - start_time) + " ms");
	}


	/**
	 * Given an ArrayList of FinchSensors and an array of sample times to query,
	 * measure the amount of time (in ms) that is required to query all of the keys.
	 * 
	 * 
	 * @param list	An ArrayList of FinchSensor objects
	 * @param keys	An array of keys to access during the test (sample times)
	 * 
	 * <dt><b>Conditions:</b>
     * <dd>PRE  -         The list and key array already exist
     * <dd>POST -         The elements of the list will be reordered
	 * </dt>
	 * 
	 * <DT><B>Algorithm: </B>
	 * <DD> Start a timer
	 * <DD> Sort the list by sample time
	 * <DD> Loop over the entire set of keys to query and ask the list to retrieve
	 * the corresponding value (using a binary search)
	 * <DD> Stop the timer.
	 * </DT>
	 * <P>
	 */
	public static void queryTest(ArrayList<FinchSensor> list, Long[] keys) {
		// Create a TimeComparator
		TimeComparator comp = new TimeComparator();
		
		// Current time is the start time
		Long start_time = System.currentTimeMillis();
		
		// Sort the list: this will allow us to perform a binary search in 
		//  the next step
		Collections.sort(list, new TimeComparator());
		
		// Loop over each key in the array and search for object in the list
		for (int i = 0; i < keys.length; ++i) {
			// Create a dummy FinchSensor to search for in the list
			//  (allows us to use Collections.binarySearch() )
			FinchSensor sensorKey = new FinchSensor(null, null, null, 0.0, keys[i]);
			
			// Perform the search
			Collections.binarySearch(list, sensorKey, comp);
		}
		
		// Current time is the end time
		Long end_time = System.currentTimeMillis();
		
		// Report the amount of time required for the entire set of queries
		System.out.println("Access Time: " + new Long(end_time - start_time) + " ms");
	}
	
	/**
	 * Given an existing map that is keyed by sample time (a Long) and stores
	 * FinchSensor values, generate a new map in which the same set of values 
	 * are stored, but are keyed by the left light sensor value (an Integer).
	 * 
	 * Because the left light sensor values are not unique across all values, 
	 * they do not form a proper set.  To account for this, the Value of the
	 * new map will be an ArrayList of FinchSenors (as opposed to a single 
	 * FinchSensor).  So: for a given left light value (e.g., 93), there are
	 * possibly a number of FinchSensor instances with that value.  These 
	 * FinchSensor objects are placed into an ArrayList, which is keyed by "93"
	 * 
	 * @param map	A map from sample times (Long) to a FinchSensor object
	 * @return		A map from left light values (Integer) to an ArrayList of 
	 * 	               FinchSensor objects
	 */
	
	public static TreeMap<Integer, ArrayList<FinchSensor>> reKey(
			Map<Long, FinchSensor> map) {
		
		// Create the new TreeMap
		TreeMap<Integer, ArrayList<FinchSensor>> newMap = 
			new TreeMap<Integer, ArrayList<FinchSensor>>();

		// Loop over all values stored in the original map
		for (FinchSensor s : map.values()) {
			// TODO: add s to the appropriate ArrayList in the new TreeMap

		}
		return newMap;
	}

	/**
	 * Compute the mean value of a some sensor value across a set of
	 * FinchSensors. The particular sensor stream is specified by the
	 * SensorComparator
	 * 
	 * @param log   A set of FinchSensors
	 * @param comp  A SensorComparator that specifies the key
	 * @return      The mean value of the keys
	 * 
	 *  <P>
	 * Algorithm: Return the sum of the channel values divided by the number of
	 * samples.
	 * <P>
	 */

	static public double mean(List<FinchSensor> log, SensorComparator comp) {
		
		double out = 0.0;

		// Loop over all objects in the list
		for (FinchSensor s : log) {
			out += comp.doubleValue(s);
		}
		return (out / log.size());
	}

	/**
	 * Return the sensor sample that is the median value of the key used to sort
	 * a set.
	 * 
	 * 
	 * @param log
	 *            A set of FinchSensors, already sorted by some key
	 * @return The sensor element in the middle of the ordered list (this is the
	 *         MEDIAN sample)
	 * 
	 * <P>
	 * Algorithm: <BR>
	 * The median sample is the middle one given that the set has already been
	 * ordered.
	 * <P>
	 * <dt><b>Conditions:</b>
	 * <dd>PRE - The log must already be instantiated and sorted according to
	 * the key
	 * <dd>POST - n/a
	 * <P>
	 */

	static public FinchSensor median(List<FinchSensor> log) {
		return (log.get(log.size() / 2));
	}

	/**
	 * For the given sensor channel (as indicated by comp), display: <dt>
	 * <dd>The mean value of this sensor channel
	 * <dd>the entire sensor sample for the minimum, median and maximum channel
	 * values</dt>
	 * <P>
	 * 
	 * @param log A valid list of Finch sensor samples
	 * @param comp The sensor channel on which to organize the list
	 * 
	 * <P>
	 * 
	 * Algorithm: <BR>
	 * 1. Sort the elements of log according to comp <BR>
	 * 2. Report the mean of the sensor value <BR>
	 * 3. Given the ordering of the sorted list, we can pull out the min, median
	 * and max sensor samples by extracting the first, middle and last elements
	 * in the list.
	 * 
	 * <P>
	 * 
	 * <dt><b>Conditions:</b> 
	 * <dd>PRE - The log must already be instantiated 
	 * <dd>POST - The order of log has changed so that
	 *            it is ordered according to the key 
	 *            </DT>
	 */
	public static void displayStats(List<FinchSensor> log, SensorComparator comp) {
		// Sort the list by a particular variable in FinchSensor: smallest to
		// largest
		Collections.sort(log, comp);

		// Name of the variable
		System.out.println("\nVariable: " + comp);

		// The mean sample
		System.out.println("Mean Value: " + mean(log, comp));

		// The minimum sample: the first one in the sorted list
		System.out.println("Min:    " + log.get(0));

		// The median sample: no loner makes sense since this is a set and not a
		// list.
		System.out.println("Median: " + median(log));

		// The max sample: the last one in the list
		System.out.println("Max:    " + log.get(log.size() - 1));

		System.out.println("");
	}

	/**
	 * Given an ArrayList of FinchSensors, report the statistics for
	 *  for both the Acceleration Magnitude and for the Z component of 
	 *  Acceleration 
	 * 
     *  
	 * @param list	ArrayList of FinchSensor objects over which to compute 
	 * 			the statistics
	 * 
	 * <dt><b>Conditions:</b>
	 * <dd>PRE  -         The ArrayList already exists and contains at 
	 * least one entry
     * <dd>POST -         The ArrayList order is altered
     * </dt>
	 */
	
	public static void displayConditionalStats(ArrayList<FinchSensor> list) {
		// Report the number of samples
		System.out.println("Sample size:" + list.size());
		
		displayStats(list, new AccelComparator());
		displayStats(list, new ZaccelComparator());

	}

	/**
	 * Create several Maps of FinchSensor objects.  We are looking to
	 *  answer several questions: <BR>
	 * 1. What order relationships do the map induce? <BR>
	 * 2. What is the relative performance of the various map implementations?<BR>
	 * 3. How can we use Maps to facilitate the computation of conditional statistics? <BR>
	 * 
	 * @param args	Are ignored
	 */
	
	public static void main(String[] args) {
		Finch myFinch = null;
		myFinch = new Finch();

		Random rand = new Random();

		// Create a HashMap:
		Map<Long, FinchSensor> hashMap = new HashMap<Long, FinchSensor>();
		
		// Sample FinchSensor objects and place them into the map
		getSamples(myFinch, numSamples, sampleInterval, hashMap);

		// Disconnect from Finch
		if(myFinch != null){
			myFinch.quit();
		}
		
		// Show the samples
		System.out.println("HASH MAP:");
		reportSamples(hashMap);

		// TODO: MILESTONE 1 QUESTION
		Map<Long, FinchSensor> treeMap = new TreeMap<Long, FinchSensor>(hashMap);  

		// Show the samples in the TreeMap
		System.out.println("\nTREE MAP:");
		reportSamples(treeMap);

		// END Milestone 1
		
		// TODO: Milestone 2: remove next line
		System.exit(0);
		
		//////////////////////////////////////////////////////////////
		// Timing test: try accessing the map many times
		Long[] keys = new Long[10000];

		// The set of keys in the map
		Set<Long> keySet = hashMap.keySet();

		// Convert this set into an array so that we can randomly sample from it
		Long[] array = new Long[keySet.size()];
		array = hashMap.keySet().toArray(array);

		// Sample N times from this set
		for (int i = 0; i < keys.length; ++i) {
			keys[i] = array[rand.nextInt(array.length)];
		}

		System.out.println("************************************");
		// Query each of the maps many times and measure the computation time required
		System.out.println("Hash map:");
		queryTest(hashMap, keys);
		
		System.out.println("\nTree map:");
		queryTest(treeMap, keys);
		
		// Convert the values contained within the map to a list 
		//  and query the list (instead of a Map)
		System.out.println("\nList:");
		ArrayList<FinchSensor> list = new ArrayList<FinchSensor>(hashMap.values());
		queryTest(list, keys);

		System.out.println("\n\n\n");
		// TODO: Milestone 3: remove next line
		System.exit(0);
		
		///////////////////////////////////////////////////////////////
		// Construct a map that is keyed using the left light sensor
		TreeMap<Integer, ArrayList<FinchSensor>> lightMap = reKey(treeMap);

		// Display statistics for the highest, middle and lowest left light sensor
		// values
		
		// Get an array that contains all of the left light sensor keys
		Integer[] lightKeys = new Integer[0];
		lightKeys = lightMap.keySet().toArray(lightKeys);
		
		System.out.println("************************************");
		System.out.println("Statistics given highest left light value:");
		displayConditionalStats(lightMap.get(lightKeys[lightKeys.length-1]));

		System.out.println("************************************");
		System.out.println("Statistics given middle left light value:");
		displayConditionalStats(lightMap.get(lightKeys[lightKeys.length/2]));
		
		System.out.println("************************************");
		System.out.println("Statistics given lowest left light value:");
		displayConditionalStats(lightMap.get(lightKeys[0]));
	}
}
